from typing import Dict, List, Optional

from ambient_backend_api_client import Service, ServiceNodeRelationship
from pydantic import BaseModel


class DockerServiceVersion(BaseModel):
    Index: int


class DockerServiceMount(BaseModel):
    Type: str
    Source: str
    Target: str


class DockerServiceContainerSpec(BaseModel):
    Image: str
    Args: List[str] = []
    Init: Optional[bool] = None
    DNSConfig: Dict[str, str] = {}
    Isolation: Optional[str] = None
    Env: List[str] = []
    Hostname: Optional[str] = None
    Mounts: List[DockerServiceMount] = []


class DockerServiceResources(BaseModel):
    Limits: Dict[str, str] = {}
    Reservations: Dict[str, str] = {}


class DockerServicePlacementPlatform(BaseModel):
    Architecture: Optional[str] = None
    OS: str


class DockerServicePlament(BaseModel):
    Platforms: List[DockerServicePlacementPlatform] = []


class DockerServiceTaskTemplate(BaseModel):
    ContainerSpec: DockerServiceContainerSpec
    Resources: Optional[DockerServiceResources] = None
    Placement: DockerServicePlament
    ForceUpdate: int
    Runtime: str


class DockerServiceModeReplicated(BaseModel):
    Replicas: int


class DockerServiceMode(BaseModel):
    Replicated: DockerServiceModeReplicated


class DockerServicePort(BaseModel):
    Protocol: str
    TargetPort: int
    PublishedPort: int
    PublishMode: str


class DockerServiceEndpointSpec(BaseModel):
    Mode: Optional[str] = None
    Ports: List[DockerServicePort] = []


class DockerServiceEndpoint(BaseModel):
    Spec: DockerServiceEndpointSpec


class DockerServiceSpec(BaseModel):
    Name: str
    Labels: Dict[str, str]
    TaskTemplate: DockerServiceTaskTemplate
    Mode: DockerServiceMode
    EndpointSpec: DockerServiceEndpointSpec


class DockerServiceAttrs(BaseModel):
    ID: str
    Version: DockerServiceVersion
    CreatedAt: str
    UpdatedAt: str
    Spec: DockerServiceSpec
    Endpoint: DockerServiceEndpoint


class ServiceNodeRelationshipWithDockerAttrs(ServiceNodeRelationship):
    class Config:
        from_attributes = True

    docker_service_attrs: Optional[DockerServiceAttrs] = None


class ServiceWithDockerAttrs(Service):
    class Config:
        from_attributes = True

    docker_service_attrs: Optional[DockerServiceAttrs] = None


class SyncSurveyReport(BaseModel):
    """For all properties, True means the property is in sync,
    False means it is not in sync"""

    both_inputs_are_not_none: bool = False
    image_in_sync: bool = False
    ports_in_sync: bool = False
    replicas_in_sync: bool = False
    labels_in_sync: bool = False
    env_vars_in_sync: bool = False
    hostname_in_sync: bool = False
    mounts_in_sync: bool = False
    networks_in_sync: bool = True  # TODO: Implement network comparison

    @property
    def in_sync(self) -> bool:
        return all(
            [
                self.both_inputs_are_not_none,
                self.image_in_sync,
                self.ports_in_sync,
                self.replicas_in_sync,
                self.labels_in_sync,
                self.env_vars_in_sync,
                self.hostname_in_sync,
                self.mounts_in_sync,
                self.networks_in_sync,
            ]
        )



from abc import ABC, abstractmethod
from typing import List, Optional, Union

import aiohttp
from ambient_backend_api_client import (
    ApiClient,
    Cluster,
    ClustersApi,
    Configuration,
    DockerClusterData,
)
from ambient_backend_api_client import NodeOutput as Node
from ambient_backend_api_client import NodeRoleEnum, NodesApi, UpdateCluster
from result import Err, Ok, Result

from ambient_client_common.models.cluster_diff import (
    ClusterDiff,
    ReconciliationPlan,
    ReconciliationStep,
)
from ambient_client_common.repositories.docker_repo import DockerRepo
from ambient_client_common.repositories.node_repo import NodeRepo
from ambient_client_common.utils import logger


class ClusterConfigService(ABC):
    @abstractmethod
    async def generate_diff(self, cluster_id: int) -> Result[ClusterDiff, str]:
        pass

    @abstractmethod
    async def plan_reconciliation(
        self, diff: ClusterDiff
    ) -> Result[ReconciliationPlan, str]:
        pass

    @abstractmethod
    async def reconcile(self, plan: ReconciliationPlan) -> Result[str, str]:
        pass


class DockerClusterConfigService(ClusterConfigService):
    def __init__(
        self, docker_repo: DockerRepo, node_repo: NodeRepo, api_config: Configuration
    ):
        self.api_config = api_config
        self.docker_repo = docker_repo
        self.node_repo = node_repo

    async def refresh_node(self) -> Node:
        current_node = self.node_repo.get_node_data()
        updated_node: Optional[Node] = None
        resp_text: Optional[str] = None
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    f"{self.api_config.host}/nodes/{current_node.id}",
                    headers={"Authorization": f"Bearer {self.api_config.access_token}"},
                ) as response:
                    resp_text = await response.text()
                    response.raise_for_status()
                    node_d = await response.json()
                    updated_node = Node.model_validate(node_d)
                    logger.debug(
                        "Node refreshed: {}", updated_node.model_dump_json(indent=4)
                    )
        except Exception as e:
            err_msg = f"Failed to refresh node: {e}"
            if resp_text:
                err_msg += f"\n{resp_text}"
            logger.error(err_msg)
            raise Exception(err_msg) from e

        if updated_node:
            self.node_repo.save_node_data(updated_node)
            return updated_node

    async def generate_diff(
        self, cluster_id: Union[int, None]
    ) -> Result[ClusterDiff, str]:
        logger.info("Generating diff for cluster: {}", cluster_id)

        # retrieve current state from device
        docker_info = self.docker_repo.get_docker_info()
        logger.debug(f"Current state: {docker_info.model_dump_json(indent=4)}")

        if cluster_id is None:
            logger.info("Desired state is absent. Cluster should not exist.")
            return Ok(ClusterDiff(current_state=docker_info.Swarm, desired_state=None))

        # Fetch the desired state from the backend
        cluster: Optional[Cluster] = None
        async with ApiClient(self.api_config) as api_client:
            clusters_api = ClustersApi(api_client)
            try:
                cluster = await clusters_api.get_cluster_clusters_cluster_id_get(
                    cluster_id
                )
                logger.debug(f"Fetched cluster: {cluster.model_dump_json(indent=4)}")
            except Exception as e:
                return Err(f"Failed to fetch cluster {cluster_id} from backend: {e}")

        return Ok(ClusterDiff(current_state=docker_info.Swarm, desired_state=cluster))

    async def plan_reconciliation(
        self, diff: ClusterDiff
    ) -> Result[ReconciliationPlan, str]:
        logger.info("Planning reconciliation ...")
        reconciliation_plan = ReconciliationPlan(
            steps=[], cluster_id=diff.desired_state.id if diff.desired_state else None
        )
        logger.debug(f"Diff: {diff.model_dump_json(indent=4)}")
        if diff.desired_state is None:
            logger.info("Desired state is absent. Cluster should not exist.")
            reconciliation_plan.steps.append(
                ReconciliationStep(
                    action=handle_cluster_should_not_exist(
                        docker_repo=self.docker_repo
                    ),
                    changes={"cluster_id": None},
                )
            )
            logger.debug(f"Reconciliation plan: {reconciliation_plan}")
            return Ok(reconciliation_plan)

        try:
            manager_nodes: List[Node] = await self._get_manager_nodes(
                diff.desired_state.id
            )
            logger.debug("retrived {} manager nodes", len(manager_nodes))
            node = await self.refresh_node()
            if self.node_repo.get_node_data().role == NodeRoleEnum.MANAGER:
                # handle cluster manager join or create
                logger.info("Node role is manager.")
                reconciliation_plan.steps.append(
                    ReconciliationStep(
                        action=handle_cluster_manager_join_or_create(
                            cluster=diff.desired_state,
                            manager_nodes=manager_nodes,
                            docker_repo=self.docker_repo,
                            api_config=self.api_config,
                            node=node,
                        ),
                        changes={"cluster_id": diff.desired_state.id},
                    )
                )
            elif self.node_repo.get_node_data().role == NodeRoleEnum.WORKER:
                # handle cluster worker join
                logger.info("Node role is worker.")
                reconciliation_plan.steps.append(
                    ReconciliationStep(
                        action=handle_cluster_worker_join(),
                        msg="Cluster worker will join cluster.",
                    )
                )
            else:
                logger.error("Node role is not set.")
                return Err("Node role is not set.")
        except Exception as e:
            logger.error(f"Failed to plan reconciliation: {e}")
            return Err(f"Failed to plan reconciliation: {e}")

        logger.debug(f"Reconciliation plan: {reconciliation_plan}")
        return Ok(reconciliation_plan)

    async def reconcile(self, plan: ReconciliationPlan) -> Result[str, str]:
        logger.info("Reconciling ...")
        return await plan.execute()

    async def _get_manager_nodes(self, cluster_id: int) -> List[Node]:
        async with ApiClient(self.api_config) as api_client:
            nodes_api = NodesApi(api_client)
            try:
                nodes_response = await nodes_api.get_nodes_nodes_get(
                    role="manager", cluster_id=cluster_id
                )
                nodes = nodes_response.results
                return nodes
            except Exception as e:
                logger.error(f"Failed to fetch manager nodes: {e}")
                return []


# handle cluster should not exist
async def handle_cluster_should_not_exist(docker_repo: DockerRepo) -> Result[str, str]:
    if docker_repo.is_node_part_of_cluster():
        docker_repo.leave_cluster()
        return Ok("Cluster left.")
    else:
        logger.info("Cluster does not exist on this node.")
        return Ok("Cluster does not exist on this node.")


async def handle_cluster_manager_join_or_create(
    cluster: Cluster,
    manager_nodes: List[Node],
    node: Node,
    docker_repo: DockerRepo,
    api_config: Configuration,
) -> Result[str, str]:
    logger.info("Handling cluster manager join or create ...")
    # is this node already part of a swarm?
    docker_info = docker_repo.get_docker_info()
    if docker_repo.is_node_part_of_cluster():
        # yes, this node is already part of a swarm
        logger.info("This node is already part of a swarm.")
        # does the cluster ID match the current cluster?
        if (
            cluster.docker_data
            and docker_info.Swarm.Cluster
            and (
                cluster.docker_data.cluster_id
                == docker_repo.get_docker_info().Swarm.Cluster.ID
            )
        ):
            # yes, the cluster ID matches the current cluster
            logger.info("Cluster ID matches the current cluster.")
            # is this node a manager?
            if docker_info.Swarm.NodeID in [
                remote_manager.NodeID
                for remote_manager in docker_info.Swarm.RemoteManagers
            ]:
                # yes, nothing to do
                logger.info("This node is already a manager in the cluster.")
                return Ok("This node is already a manager in the cluster.")
            else:
                # no, this node is not a manager
                # leave cluster and join as manager
                logger.info(
                    "This node is not a manager in the cluster. Leaving cluster ..."
                )
                docker_repo.leave_cluster()
                logger.info("Cluster left. Joining cluster as a manager ...")
                return await join_cluster(
                    NodeRoleEnum.MANAGER, manager_nodes, docker_repo
                )
        else:
            # no, the cluster ID does not match the current cluster
            # leave cluster and join as manager
            logger.info(
                "Cluster ID does not match the current cluster. Leaving cluster ..."
            )
            docker_repo.leave_cluster()
            logger.info("Cluster left. Joining cluster as a manager ...")
            # make sure that this isn't the first node in the cluster
            if manager_nodes != [node]:
                logger.info("This is not the first node in the cluster.")
                logger.debug("Manager nodes: {}", manager_nodes)
                logger.debug("Current node: {}", [node])
                return await join_cluster(
                    NodeRoleEnum.MANAGER, manager_nodes, docker_repo
                )
            logger.info("This is the first node in the cluster.")

    # no, this node is not part of a swarm
    logger.info("This node is not part of a swarm.")
    # has the cluster been initiated?

    if len(manager_nodes) > 1:
        # yes, the cluster has been initiated
        # join as manager
        logger.info("Cluster has been initiated. Joining cluster as a manager ...")
        return await join_cluster(NodeRoleEnum.MANAGER, manager_nodes, docker_repo)
    else:
        # no, the cluster has not been initiated
        # create cluster
        logger.info("Cluster has not been initiated. Creating cluster ...")
        ad_addr_if = next(
            (
                i.ipv4_address
                for i in node.interfaces
                if i.name.startswith(("eth", "wlan"))
            ),
            None,
        )
        if not ad_addr_if:
            logger.error("Failed to get advertise address.")
            return Err("Failed to get advertise address.")
        docker_repo.create_cluster(ad_addr_if)
        logger.info("Cluster created.")
        # update cluster initiated status
        docker_info = docker_repo.get_docker_info()
        updated_cluster_data = UpdateCluster.model_validate(cluster.model_dump())
        updated_cluster_data.docker_data = DockerClusterData(
            initiated=True,
            cluster_id=docker_info.Swarm.Cluster.ID,
            remote_managers=[
                rm.model_dump() for rm in docker_info.Swarm.RemoteManagers
            ],
        )
        return await update_cluster_docker_data(
            updated_cluster_data, cluster.id, api_config
        )


async def join_cluster(
    role: NodeRoleEnum, manager_nodes: List[Node], docker_repo: DockerRepo
) -> Result[str, str]:
    logger.info("Joining cluster as a manager ...")
    # cycle through manager nodes until we successfully join the cluster
    join_token: Optional[str] = None
    for manager_node in manager_nodes:
        logger.info(
            "Getting join token from manager node: {} - {}",
            manager_node.name,
            manager_node.docker_swarm_info.node_addr,
        )
        join_token = await get_join_token(
            manager_node.docker_swarm_info.node_addr, role.value
        )
        if join_token:
            logger.info("Join token received: {}", join_token)
            break
    if join_token:
        logger.info("Joining cluster ...")
        if docker_repo.join_cluster(
            remote_addrs=[node.docker_swarm_info.node_addr for node in manager_nodes],
            join_token=join_token,
        ):
            logger.info("Cluster joined successfully.")
            return Ok("Cluster joined successfully.")
        else:
            logger.error("Failed to join cluster.")
            return Err("Failed to join cluster.")
    else:
        logger.error("Failed to get join token.")
        return Err("Failed to get join token.")


async def handle_cluster_worker_join(
    cluster: Cluster,
    manager_nodes: List[Node],
    docker_repo: DockerRepo,
    api_config: Configuration,
) -> Result[str, str]:
    # is this node already part of a swarm?
    docker_info = docker_repo.get_docker_info()
    if docker_repo.is_node_part_of_cluster():
        # yes, this node is already part of a swarm
        # does the cluster ID match the current cluster?
        if cluster.docker_data.cluster_id == docker_info.Swarm.Cluster.ID:
            # yes, the cluster ID matches the current cluster
            # is this node a worker?
            if docker_info.Swarm.NodeID not in [
                manager.NodeID for manager in docker_info.Swarm.RemoteManagers
            ]:
                # yes, nothing to do
                logger.info("This node is already a worker in the cluster.")
                return Ok("This node is already a worker in the cluster.")
            else:
                # no, this node is not a worker
                # leave cluster and join as worker
                logger.info(
                    "This node is not a worker in the cluster. Leaving cluster ..."
                )
                docker_repo.leave_cluster()
                logger.info("Cluster left. Joining cluster as a worker ...")
                return await join_cluster(
                    NodeRoleEnum.WORKER, manager_nodes, docker_repo
                )
        else:
            # no, the cluster ID does not match the current cluster
            # leave cluster and join as worker
            logger.info(
                "Cluster ID does not match the current cluster. Leaving cluster ..."
            )
            docker_repo.leave_cluster()
            logger.info("Cluster left. Joining cluster as a worker ...")
            return await join_cluster(NodeRoleEnum.WORKER, manager_nodes, docker_repo)
    else:
        # no, this node is not part of a swarm
        # join as worker
        logger.info("Joining cluster as a worker ...")
        return await join_cluster(NodeRoleEnum.WORKER, manager_nodes, docker_repo)


async def get_join_token(manager_addr: str, role: str) -> str:
    async with aiohttp.ClientSession() as session:
        async with session.get(
            f"http://{manager_addr}/data/swarm/join-token?role={role}"
        ) as response:
            response.raise_for_status()
            data = await response.json()
            logger.debug("Join token response: {}", data)
            return data["join_token"]


async def update_cluster_docker_data(
    data: UpdateCluster, cluster_id: int, api_config: Configuration
) -> Result[str, str]:
    logger.info("Patching cluster with data: {}", data.model_dump_json(indent=4))
    logger.debug("Cluster ID: {}", cluster_id)

    try:
        async with ApiClient(api_config) as api_client:
            clusters_api = ClustersApi(api_client)
            cluster = await clusters_api.update_cluster_clusters_cluster_id_put(
                cluster_id, data
            )
            logger.debug("Cluster patched: {}", cluster.model_dump_json(indent=4))
            return Ok("Cluster patched successfully")
    except Exception as e:
        err_msg = f"Failed to PUT cluster: {e}"
        logger.error(err_msg)
        return Err(err_msg)



from logging import Logger
from typing import Any, Optional, Union

import aiohttp
from ambient_backend_api_client import ApiClient, Cluster, ClustersApi, Configuration
from ambient_backend_api_client import NodeOutput as Node
from ambient_base_plugin import BasePlugin, hookimpl
from ambient_base_plugin.models.configuration import ConfigPayload
from ambient_base_plugin.models.message import Message
from docker_swarm_plugin.services.cluster_config_service import (
    ClusterConfigService,
    DockerClusterConfigService,
)
from docker_swarm_plugin.services.service_config_service import ServiceConfigSvc
from fastapi import HTTPException
from pydantic import ValidationError

from ambient_client_common import config
from ambient_client_common.models.docker_models import SwarmAttrs
from ambient_client_common.repositories.docker_repo import DockerRepo
from ambient_client_common.repositories.node_repo import NodeRepo

logger: Optional[Logger] = None


class DockerSwarmPlugin(BasePlugin):
    def __init__(self) -> None:
        self.node_repo: Optional[NodeRepo] = None
        self.docker_repo: Optional[DockerRepo] = None
        self.api_config: Optional[Configuration] = None
        self.logger: Optional[Logger] = None
        self.cluster_config_service: Optional[ClusterConfigService] = None
        self.service_config_svc: Optional[ServiceConfigSvc] = None

    @hookimpl
    async def run_system_sweep(self, node: Node) -> None:
        logger.info("Running system sweep from the Docker Swarm Plugin ...")
        if node.role.value == "worker":
            logger.info("Cannot run docker swarm system sweep on worker node")
            return
        await self.service_config_svc.run_system_sweep(node)

    async def configure(
        self, config: ConfigPayload, logger: Union[Logger, Any] = None
    ) -> None:
        self.node_repo = config.node_repo
        self.docker_repo = config.docker_repo
        self.api_config = config.api_config
        self.set_logger(logger)
        self.cluster_config_service = DockerClusterConfigService(
            docker_repo=self.docker_repo,
            node_repo=self.node_repo,
            api_config=self.api_config,
        )
        self.service_config_svc = ServiceConfigSvc(
            docker_repo=self.docker_repo, config_payload=config
        )
        logger.info("Configured DockerSwarmPlugin")

    def set_logger(self, logger_: Logger) -> None:
        global logger
        logger = logger_
        self.logger = logger_

    async def handle_event(self, message: Message) -> None:
        """Handle incoming messages

        Args:
            message (Message): Incoming message

        Returns:
            None
        """
        logger.info("Handling event [from docker swarm plugin handler] ...")
        if "CLUSTER_EVENT" in message.topic:
            return await self.handle_cluster_event(message)
        elif "SERVICE_EVENT" in message.topic:
            service_id = int(message.topic.split("/")[-2])
            logger.info(f"Handling service event for service: {service_id}")
            return await self.service_config_svc.handle_service_event(
                service_id=service_id
            )
        else:
            logger.error(f"Unsupported message topic: {message.topic}")

    async def handle_cluster_event(self, message: Message) -> None:
        """Handle cluster event

        Args:
            message (Message): Incoming message

        Returns:
            None
        """
        logger.info("Handling cluster event ...")
        logger.debug(f"Handling cluster event: {message.model_dump_json(indent=4)}")
        await self.reconcile_cluster()

    async def reconcile_cluster(self) -> None:
        # fetch node from from backend
        try:
            current_node_data: Node = self.node_repo.get_node_data(strict=True)
            logger.info("retrieved node data from local repo")
            fetched_node_data = await fetch_node_data(
                current_node_data.id, self.api_config
            )
            logger.info("fetched node data from backend")
            self.node_repo.save_node_data(fetched_node_data)
            logger.debug(
                f"Fetched node data: {fetched_node_data.model_dump_json(indent=4)}"
            )
        except Exception as e:
            logger.error(f"Failed to fetch node data: {e}")
            return

        try:
            # create diff
            diff_result = await self.cluster_config_service.generate_diff(
                cluster_id=fetched_node_data.cluster_id
            )
            if diff_result.is_err():
                logger.error(f"Failed to generate diff: {diff_result.unwrap_err()}")
                return
            diff = diff_result.unwrap()
            logger.debug(f"Generated diff: {diff.model_dump_json(indent=4)}")

            # generate plan
            logger.debug("Generating reconciliation plan ...")
            plan_result = await self.cluster_config_service.plan_reconciliation(diff)
            if plan_result.is_err():
                logger.error(f"Failed to generate plan: {plan_result.unwrap_err()}")
                return
            plan = plan_result.unwrap()
            logger.debug(f"Generated plan: {plan}")

            # execute plan
            logger.debug("Executing reconciliation plan ...")
            await self.cluster_config_service.reconcile(plan=plan)
        except Exception as e:
            logger.error(f"Failed to reconcile cluster: {e}")
            return

    async def handle_api_request(
        self, method: str, path: str, data: Optional[None] = None
    ) -> Any:
        logger.info("Handling API request ...")
        method = method.lower()
        if method.upper() == "POST" and "onboarding" in path:
            logger.info("Handling onboarding request ...")
            return await self.handle_onboarding_request()
        elif method.upper() == "GET" and "swarm_info" in path:
            return await self.get_swarm_info()
        logger.error("Unsupported method. Method: {}, Path: {}", method, path)
        raise HTTPException(status_code=405, detail="Method not allowed")

    async def handle_onboarding_request(self) -> None:
        await self.reconcile_cluster()

    async def get_swarm_info(self) -> SwarmAttrs:
        logger.info("Getting swarm info ...")
        try:
            swarm_info = self.docker_repo.get_swarm_info()
        except ValidationError as e:
            logger.error(f"Failed to get swarm info: {e}")
            raise HTTPException(status_code=400, detail="Failed to get swarm info")
        return swarm_info


async def fetch_node_data(node_id: int, api_config: Configuration) -> Node:
    logger.info("Fetching node data from backend [Node ID: {}]", node_id)
    logger.debug("api_config token: {}", api_config.access_token)
    async with aiohttp.ClientSession() as session:
        async with session.get(
            f"{config.settings.backend_api_url}/nodes/{node_id}",
            headers={
                "Authorization": f"Bearer {api_config.access_token}",
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        ) as response:
            response.raise_for_status()
            node_data = await response.json()
            return Node.model_validate(node_data)


async def fetch_cluster_data(cluster_id: str, api_config: Configuration) -> Cluster:
    async with ApiClient(api_config) as api_client:
        clusters_api = ClustersApi(api_client)
        cluster = await clusters_api.get_cluster_clusters_cluster_id_get(cluster_id)
        return cluster



import asyncio
import json
from typing import Dict, List, Optional, Tuple, Union

import aiohttp
import async_lru
import docker
import docker.types
from ambient_backend_api_client import NodeOutput as Node
from ambient_backend_api_client import RequestedServiceSpec, ServiceCreate
from ambient_base_plugin.models.configuration import ConfigPayload
from docker_swarm_plugin.models.service_models import DockerServiceAttrs
from docker_swarm_plugin.models.service_models import (
    ServiceNodeRelationshipWithDockerAttrs as ServiceNodeRelationship,
)
from docker_swarm_plugin.models.service_models import ServiceWithDockerAttrs as Service
from docker_swarm_plugin.models.service_models import SyncSurveyReport
from pydantic import ValidationError
from result import Err, Ok, Result

from ambient_client_common.repositories.docker_repo import DockerRepo
from ambient_client_common.utils import logger

Pair = Tuple[Optional[DockerServiceAttrs], Optional[ServiceNodeRelationship]]


class ServiceConfigSvc:
    def __init__(self, docker_repo: DockerRepo, config_payload: ConfigPayload) -> None:
        self.docker_client: docker.DockerClient = docker_repo.client
        self.config_payload = config_payload
        self.api_config = config_payload.api_config

    async def run_system_sweep(self, node: Node) -> None:
        """Run a system sweep on the given node.
        Running a system sweep in this context is to
        check the current status of services on the node
        and update the status of the services via the API.

        Args:
            node (Node): Node object.
        """
        logger.info("Running system sweep on node: {}", node.id)
        # gather local service data
        local_services = gather_local_services(self.docker_client)
        logger.info("Gathered {} local services", len(local_services))

        # gather services from API
        logger.debug("Getting token ...")
        token = await self.config_payload.get_token()
        logger.info("Retrieved token. [{} chars]", len(token))
        logger.info("token from get_token: {}", token)
        backend_services, parent_services = (
            await get_service_node_relationships_from_api(
                node_id=node.id, base_url=self.api_config.host, token=token
            )
        )
        logger.info("Gathered {} backend services", len(backend_services))

        # compare and diff services
        logger.info("Pairing services ...")
        service_pairs = pair_services(local_services, backend_services, parent_services)
        logger.debug("Service pairs: {}", service_pairs)
        grouped_pairs = group_services(service_pairs)
        logger.debug("Grouped pairs: {}", grouped_pairs)
        logger.info(
            "Total services: {}, grouped pairs: {}",
            len(service_pairs),
            sum(len(v) for v in grouped_pairs.values()),
        )

        # update service data via API
        await handle_paired(
            pairs=grouped_pairs["paired"],
            base_url=self.api_config.host,
            token=self.api_config.access_token,
            node_id=node.id,
        )
        logger.info("Handled paired services")

        # create out of sync services
        await handle_local_only(
            pairs=grouped_pairs["local_only"],
            node_id=node.id,
            base_url=self.api_config.host,
            token=self.api_config.access_token,
        )
        logger.info("Handled local only services")

        # mark out of sync services
        await handle_backend_only(
            pairs=grouped_pairs["backend_only"],
            base_url=self.api_config.host,
            token=self.api_config.access_token,
        )
        logger.info("Handled backend only services")

        logger.info("ServiceConfigSvc system sweep complete")

    async def handle_service_event(self, service_id: int) -> None:
        """Handle a service event."""

        logger.info("Handling service event for service: {}", service_id)

        token = await self.config_payload.get_token()
        logger.debug("Retrieved token. [{} chars]", len(token))

        # get service from API
        try:
            backend_service = await get_service(
                service_id=service_id,
                base_url=self.api_config.host,
                token=token,
            )
            if not backend_service:
                # hande error
                return await handle_service_event_error(
                    base_url=self.api_config.host,
                    token=token,
                    service_id=service_id,
                    node_id=self.config_payload.node_id,
                    error="Failed to get service from API",
                    local_service=None,
                )
            logger.info(
                "Retrieved service from API: {} ... \
[{} total fields, {} populated fields]",
                backend_service.model_dump_json()[:20],
                len(backend_service.model_dump().keys()),
                len(backend_service.model_dump(exclude_unset=True).keys()),
            )

            # get deployment
            deployment = await get_service_node_relationship(
                base_url=self.api_config.host,
                service_id=service_id,
                node_id=self.config_payload.node_id,
                token=token,
            )
            if not deployment:
                # hande error
                err_msg = "Failed to get service node relationship from API"
                logger.error(err_msg)
                return await handle_service_event_error(
                    base_url=self.api_config.host,
                    token=token,
                    service_id=service_id,
                    node_id=self.config_payload.node_id,
                    error=err_msg,
                    local_service=None,
                    existing_deployment=None,
                )
            logger.info(
                "Retrieved deployment from API: {} ... \
[{} total fields, {} populated fields]",
                deployment.model_dump_json()[:20],
                len(deployment.model_dump().keys()),
                len(deployment.model_dump(exclude_unset=True).keys()),
            )
        except Exception as e:
            logger.error("Failed to get service and deployment: {}", e)
            return await handle_service_event_error(
                base_url=self.api_config.host,
                token=token,
                service_id=service_id,
                node_id=self.config_payload.node_id,
                error="Failed to get service and deployment",
                local_service=None,
            )

        # find matching service in local services
        local_services = gather_local_services(self.docker_client)
        logger.info("Found {} local services", len(local_services))
        matching_local_service = find_matching_local_service(
            backend_service=deployment,
            parent_service=backend_service,
            local_services=local_services,
        )

        updated_local_service: Optional[DockerServiceAttrs] = None
        try:
            # if match, enforce requested spec
            if matching_local_service:
                logger.info(
                    "Found matching local service: {}", matching_local_service.Spec.Name
                )
                updated_local_service = enforce_requested_spec(
                    matching_local_service,
                    deployment,
                    backend_service,
                    client=self.docker_client,
                )
                if not updated_local_service:
                    # hande error
                    err_msg = "Failed to enforce requested spec"
                    logger.error(err_msg)
                    return await handle_service_event_error(
                        base_url=self.api_config.host,
                        token=token,
                        service_id=service_id,
                        node_id=self.config_payload.node_id,
                        error=err_msg,
                        local_service=None,
                        existing_deployment=deployment,
                    )
                logger.info(
                    "Enforced requested spec on local service: {}",
                    updated_local_service.Spec.Name,
                )

            # if no match, create service
            else:
                updated_local_service_result = create_local_service(
                    parent_service=backend_service, client=self.docker_client
                )
                if updated_local_service_result.is_err():
                    # handle error
                    err_msg = (
                        "Failed to create local service: "
                        + updated_local_service_result.unwrap_err()
                    )
                    logger.error(err_msg)
                    return await handle_service_event_error(
                        base_url=self.api_config.host,
                        token=token,
                        service_id=service_id,
                        node_id=self.config_payload.node_id,
                        error=err_msg,
                        local_service=None,
                        existing_deployment=deployment,
                    )
                updated_local_service = updated_local_service_result.unwrap()
                logger.info(
                    "Created local service: {}", updated_local_service.Spec.Name
                )
        except Exception as e:
            err_msg = f"Failed to enforce requested spec or create local service: {e}"
            logger.error(err_msg)
            return await handle_service_event_error(
                base_url=self.api_config.host,
                token=token,
                service_id=service_id,
                node_id=self.config_payload.node_id,
                error=err_msg,
                local_service=None,
                existing_deployment=deployment,
            )

        # handle success
        return await handle_service_event_success(
            base_url=self.api_config.host,
            token=token,
            service_id=service_id,
            node_id=self.config_payload.node_id,
            local_service=updated_local_service,
        )


async def handle_service_event_error(
    base_url: str,
    token: str,
    service_id: int,
    node_id: int,
    error: str,
    local_service: Optional[DockerServiceAttrs],
    existing_deployment: Optional[ServiceNodeRelationship] = None,
) -> None:
    # update service deployment
    logger.info("Handling service event error ...")
    try:
        updated_deployment = ServiceNodeRelationship(
            service_id=service_id,
            node_id=node_id,
            state=existing_deployment.state if existing_deployment else None,
            status="failure",
            error=error,
            docker_service_attrs=local_service,
        )
        logger.debug(
            "Updated deployment: {}", updated_deployment.model_dump_json(indent=4)
        )
        return await update_service_node_relationship(
            base_url=base_url,
            token=token,
            service_node_relationship=updated_deployment,
        )
    except Exception as e:
        logger.error("Failed to handle service event error: {}", e)


async def handle_service_event_success(
    base_url: str,
    token: str,
    service_id: int,
    node_id: int,
    local_service: DockerServiceAttrs,
) -> None:
    logger.info("Handling service event success ...")
    # update service deployment
    updated_deployment = ServiceNodeRelationship(
        service_id=service_id,
        node_id=node_id,
        state="deployed",
        status="success",
        docker_service_attrs=local_service,
    )
    return await update_service_node_relationship(
        base_url=base_url,
        token=token,
        service_node_relationship=updated_deployment,
    )


def create_local_service(
    parent_service: Service, client: docker.DockerClient
) -> Result[Optional[DockerServiceAttrs], str]:
    """Create a local service from the backend service."""
    logger.info(
        "Creating service using docker client. Service name: {}, image: {}",
        parent_service.name,
        parent_service.requested_service_spec.image,
    )
    try:
        labels = parent_service.requested_service_spec.labels
        if labels:
            requested_labels_dict = {
                label.split("=")[0]: label.split("=")[1]
                for label in parent_service.requested_service_spec.labels
            }
        else:
            requested_labels_dict = {}
        requested_spec = parent_service.requested_service_spec
        docker_service = client.services.create(
            image=requested_spec.image,
            maxreplicas=requested_spec.replicas,
            container_labels=requested_labels_dict,
            endpoint_spec=build_endpoint_spec(parent_service),
            env=requested_spec.env_vars,
            hostname=requested_spec.hostname if requested_spec.hostname else None,
            labels=requested_labels_dict,
            mounts=requested_spec.mounts,
            name=parent_service.name,
            networks=requested_spec.networks,
        )
        return Ok(DockerServiceAttrs.model_validate(docker_service.attrs))
    except docker.errors.APIError as e:
        logger.error("Encountered API error: {}", e)
        # run update if error code 409
        if e.response.status_code == 409:
            logger.info("Service already exists. Updating service ...")
            existing_service = get_local_service(client, parent_service.name)
            if not existing_service:
                logger.error("Failed to get existing service")
                return Err("Failed to get existing service" + str(e))
            return Ok(
                update_local_service(
                    local_service=existing_service,
                    parent_service=parent_service,
                    client=client,
                )
            )
        return Err(str(e))
    except Exception as e:
        logger.error("Failed to create service using docker client: {}", e)
        return Err(str(e))


def get_local_service(
    client: docker.DockerClient, name: str
) -> Optional[DockerServiceAttrs]:
    """Get a local service from the given Docker client."""
    logger.info("Getting local service ...")
    try:
        service = client.services.list(filters={"name": name})[0]
        return DockerServiceAttrs.model_validate(service.attrs)
    except Exception as e:
        logger.error("Failed to get local service: {}", e)
        return None


def update_local_service(
    local_service: DockerServiceAttrs,
    parent_service: Service,
    client: docker.DockerClient,
) -> Optional[DockerServiceAttrs]:
    """Update a local service from the backend service."""
    logger.info(
        "Updating service using docker client. Service name: {}, image: {}",
        parent_service.name,
        parent_service.requested_service_spec.image,
    )
    try:
        labels = parent_service.requested_service_spec.labels
        if labels:
            requested_labels_dict = {
                label.split("=")[0]: label.split("=")[1]
                for label in parent_service.requested_service_spec.labels
            }
        else:
            requested_labels_dict = {}
        requested_spec = parent_service.requested_service_spec
        existing_service = client.services.get(local_service.ID)
        docker_service = existing_service.update(
            image=requested_spec.image,
            maxreplicas=requested_spec.replicas,
            container_labels=requested_labels_dict,
            endpoint_spec=build_endpoint_spec(parent_service),
            env=requested_spec.env_vars,
            hostname=requested_spec.hostname if requested_spec.hostname else None,
            labels=requested_labels_dict,
            mounts=requested_spec.mounts,
            name=parent_service.name,
            networks=requested_spec.networks,
        )
        logger.info("new service: {}", docker_service)
        logger.info("Service updated. Reloading service ...")
        existing_service.reload()
        return DockerServiceAttrs.model_validate(existing_service.attrs)
    except Exception as e:
        logger.error("Failed to update service using docker client: {}", e)
        return None


def build_endpoint_spec(parent_service: Service) -> Optional[docker.types.EndpointSpec]:
    """Build enpoint spec for creating a service locally"""
    if not parent_service.requested_service_spec.ports:
        return None

    logger.info("building endpoint spec ...")
    ports_dict = {}  # {publish_port: target_port}
    for ports in parent_service.requested_service_spec.ports:
        port_list = ports.split(":")
        if len(port_list) != 2:
            logger.error("Invalid port format: {}", ports)
            continue
        ports_dict[int(port_list[0])] = int(port_list[1])

    logger.info("ports_dict: {}", ports_dict)

    endpoint_spec = docker.types.EndpointSpec(
        mode="vip",
        ports=ports_dict,
    )
    logger.info("endpoint_spec: {}", endpoint_spec)
    return endpoint_spec


def enforce_requested_spec(
    local_service: DockerServiceAttrs,
    backend_service: ServiceNodeRelationship,
    parent_service: Service,
    client: docker.DockerClient,
) -> DockerServiceAttrs:
    """Enforce the requested spec on the local service."""
    sync_report = determine_in_sync(local_service, backend_service, parent_service)
    if sync_report.in_sync:
        logger.info("Service is in sync. No need to enforce requested spec.")
        return local_service
    else:
        logger.info("Service is out of sync. Need to enfore requested spec ...")
        return update_local_service(local_service, parent_service, client)


async def get_service_node_relationship(
    base_url: str, service_id: int, node_id: int, token: str
) -> Optional[ServiceNodeRelationship]:
    """Get the service node relationship."""
    logger.info("Getting service node relationship ...")
    url = f"{base_url}/services/{service_id}/node_relationships/{node_id}/"
    headers = {"Authorization": f"Bearer {token}"}
    async with aiohttp.ClientSession() as session:
        try:
            resp, resp_text = await make_api_request(session, url, headers)
            resp.raise_for_status()
            return ServiceNodeRelationship.model_validate_json(resp_text)
        except Exception as e:
            logger.error("Failed to get service node relationship: {}", e)
            return None


@async_lru.alru_cache(maxsize=128, ttl=60)
async def get_service_cached(
    service_id: int, base_url: str, token: str
) -> Optional[Service]:
    """Get the service from the cache."""
    return await get_service(service_id, base_url, token)


async def get_service(service_id: int, base_url: str, token: str) -> Optional[Service]:
    logger.info("Getting service ...")
    resp_text: Optional[str] = None
    url = f"{base_url}/services/{service_id}/"
    headers = {"Authorization": f"Bearer {token}"}
    logger.debug("Request full URL: GET - {}", url)
    logger.debug("Headers: {}", headers)
    async with aiohttp.ClientSession() as session:
        try:
            resp, resp_text = await make_api_request(session, url, headers)
            logger.debug("Response: {}", resp_text)
            resp.raise_for_status()
            return Service.model_validate_json(resp_text)
        except Exception as e:
            logger.error(
                "Failed to get service: {}",
                (str(e) + "\n" + resp_text) if resp_text else e,
            )
            return None


def find_matching_local_service(
    backend_service: ServiceNodeRelationship,
    parent_service: Service,
    local_services: List[DockerServiceAttrs],
) -> Optional[DockerServiceAttrs]:
    """Find the matching local service."""
    logger.info("Finding matching local service ...")
    try:
        for local_service in local_services:
            logger.debug(
                "Comparing local service: {} with backend service: {} ...",
                local_service.Spec.Name,
                parent_service.name,
            )
            if match_service(local_service, backend_service, parent_service):
                logger.info("Found matching local service: {}", local_service.Spec.Name)
                return local_service
        logger.info("No matching local service found")
        return None
    except Exception as e:
        logger.error("Failed to find matching local service: {}", e)
        return None


def gather_local_services(client: docker.DockerClient) -> List[DockerServiceAttrs]:
    """Gather the current services from the given Docker client.

    Args:
        client (docker.DockerClient): The Docker client to use.

    Returns:
        List[DockerServiceAttrs]: The list of current services.
    """
    logger.info("Gathering current services ...")
    services = client.services.list()
    logger.info("Collected {} services", len(services))
    # for service in services:
    #     logger.debug("service data: {}", json.dumps(service.attrs, indent=4))

    # return [DockerServiceAttrs.model_validate(service.attrs) for service in services]
    parsed_models = []
    for service in services:
        try:
            parsed_models.append(DockerServiceAttrs.model_validate(service.attrs))
        except ValidationError as e:
            logger.error("Failed to parse service: {}", e)
    return parsed_models


async def make_api_request(
    session: aiohttp.ClientSession,
    url: str,
    headers: dict,
    method: str = "get",
    data: Union[dict, None] = None,
) -> Tuple[aiohttp.ClientResponse, str]:
    max_redirects = 10
    for i in range(max_redirects):
        logger.info("Fetching URL: {} [redirects: {}]", url, i)
        method_call = getattr(session, method)
        resp: aiohttp.ClientResponse
        if data:
            async with method_call(
                url, headers=headers, allow_redirects=False, json=data
            ) as resp:
                if resp.status in (301, 302, 303, 307, 308):
                    url = resp.headers["Location"]
                    logger.info("Redirecting to: {}", url)
                    continue
                else:
                    resp_text = await resp.text()
                    return resp, resp_text
        else:
            async with method_call(url, headers=headers, allow_redirects=False) as resp:
                if resp.status in (301, 302, 303, 307, 308):
                    url = resp.headers["Location"]
                    logger.info("Redirecting to: {}", url)
                    continue
                else:
                    resp_text = await resp.text()
                    return resp, resp_text
    raise Exception("Too many redirects")


async def get_services_from_api(
    node_id: int, base_url: str, token: str
) -> List[Service]:
    logger.info("Getting services from API ...")
    resp_text: Optional[str] = None
    url = f"{base_url}/nodes/{node_id}/services/"
    headers = {"Authorization": f"Bearer {token}"}
    logger.debug("Request full URL: GET - {}", url)
    logger.debug("Headers: {}", headers)
    async with aiohttp.ClientSession() as session:
        try:
            resp, resp_text = await make_api_request(session, url, headers)
            logger.debug("Response: {}", resp_text)
            resp.raise_for_status()
            data: dict = await resp.json()
            count = data.get("count", 0)
            logger.info("Got {} services from API", count)
            services = data.get("results", [])
            return [Service.model_validate(service) for service in services]
        except Exception as e:
            logger.error(
                "Failed to get services from API: {}",
                (str(e) + "\n" + resp_text) if resp_text else e,
            )
            return []


async def get_service_node_relationships_from_api(
    node_id: int, base_url: str, token: str
) -> Tuple[List[ServiceNodeRelationship], Dict[int, Service]]:
    services = await get_services_from_api(node_id, base_url, token)
    awaitables = [
        get_service_node_relationship_from_api(service.id, node_id, base_url, token)
        for service in services
    ]
    return (await asyncio.gather(*awaitables)), {
        service.id: service for service in services
    }


async def get_service_node_relationship_from_api(
    service_id: int, node_id: int, base_url: str, token: str
) -> ServiceNodeRelationship:
    logger.info("Getting service node relationship from API ...")
    resp_text: Optional[str] = None
    url = f"{base_url}/services/{service_id}/node_relationships/{node_id}/"
    headers = {"Authorization": f"Bearer {token}"}
    logger.debug("Request full URL: GET - {}", url)
    logger.debug("Headers: {}", headers)
    async with aiohttp.ClientSession() as session:
        try:
            resp, resp_text = await make_api_request(session, url, headers)
            logger.debug("Response: {}", resp_text)
            resp.raise_for_status()
            return ServiceNodeRelationship.model_validate_json(resp_text)
        except Exception as e:
            logger.error(
                "Failed to get service node relationship from API: {}",
                (str(e) + "\n" + resp_text) if resp_text else e,
            )


def match_service(
    local_service: DockerServiceAttrs,
    backend_service: ServiceNodeRelationship,
    parent_service: Service,
) -> bool:
    if not backend_service.docker_service_attrs:
        return False
    local_name = local_service.Spec.Name
    backend_name = parent_service.name
    logger.debug(
        "Matching local service: {} with backend service: {}",
        local_name,
        backend_name,
    )
    return local_name == backend_name


def pair_services(
    local_services: List[DockerServiceAttrs],
    backend_services: List[ServiceNodeRelationship],
    parent_services: Dict[int, Service],
) -> List[Pair]:
    """Pair the local services with the backend services.

    Args:
        local_services (List[DockerServiceAttrs]): The local services.
        backend_services (List[ServiceNodeRelationship]): The backend
            service node relationships.

    Returns:
        List[Pair]: The paired services.
            [
                (local_service, backend_service),
                (local_service, None),
                (None, backend_service),
                ...
            ]
    """

    logger.info("Pairing services ...")
    pairs = []
    seen_ids = set()
    seen_local_names = set()

    for local_service in local_services:
        logger.debug("Pairing local service: {}", local_service.Spec.Name)
        for backend_service in backend_services:
            parent_service = parent_services.get(backend_service.service_id)
            if not parent_service:
                logger.error(
                    "Parent service not found for service: {}",
                    backend_service.service_id,
                )
                continue
            matched_service = match_service(
                local_service, backend_service, parent_service
            )
            if matched_service:
                logger.debug("Matched service: {}", backend_service)
                pairs.append((local_service, backend_service))
                seen_ids.add(backend_service.service_id)
                seen_local_names.add(local_service.Spec.Name)
                break
        if local_service.Spec.Name not in seen_local_names:
            logger.debug(
                "No match found for local service: {}", local_service.Spec.Name
            )
            pairs.append((local_service, None))

    for backend_service in backend_services:
        if backend_service.service_id not in seen_ids:
            logger.debug(
                "No match found for backend service: {}", backend_service.service_id
            )
            pairs.append((None, backend_service))

    logger.info("Parsed {} pairs", len(pairs))
    return pairs


def group_services(pairs: List[Pair]) -> Dict[str, List[Pair]]:
    """Group the pairs of services by their name.

    Args:
        pairs (List[Pair]): The pairs of services.

    Returns:
        Dict[str, List[Pair]]: The grouped pairs.
    """
    logger.info("Grouping services ...")
    grouped = {"paired": [], "local_only": [], "backend_only": []}

    for local_service, backend_service in pairs:
        if local_service and backend_service:
            grouped["paired"].append((local_service, backend_service))
        elif local_service:
            grouped["local_only"].append((local_service, None))
        elif backend_service:
            grouped["backend_only"].append((None, backend_service))

    logger.info(
        "Grouped services: {} paired, {} local only, {} backend only",
        len(grouped["paired"]),
        len(grouped["local_only"]),
        len(grouped["backend_only"]),
    )

    return grouped


async def handle_paired(
    pairs: List[Pair], base_url: str, token: str, node_id: int
) -> None:
    """Handle the paired services.

    Args:
        pairs (List[Pair]): The paired services.
    """

    logger.info("Handling paired services ...")

    # mismatched in this context means the services are not in sync
    mismatched_pairs = []
    matched_pairs = []

    for local_service, backend_service in pairs:
        parent_service = await get_service_cached(
            backend_service.service_id, base_url, token
        )
        if determine_in_sync(local_service, backend_service, parent_service).in_sync:
            matched_pairs.append((local_service, backend_service))
        else:
            mismatched_pairs.append((local_service, backend_service))

    logger.info("In-sync: {} services", len(matched_pairs))
    logger.info("Out-of-sync: {} services", len(mismatched_pairs))

    await handle_paired_mismatched(mismatched_pairs, base_url, token, node_id)
    await handle_paired_and_in_sync(matched_pairs, base_url, token)

    logger.info("Handled paired services.")


def determine_in_sync(
    local_service: DockerServiceAttrs,
    backend_service: ServiceNodeRelationship,
    parent_service: Service,
) -> SyncSurveyReport:
    """Determine if the local service matches the backend service

    Args:
        local_service (DockerServiceAttrs): local service
        backend_service (ServiceNodeRelationship): backend service node relationship

    Returns:
        bool: True if the services match, False otherwise
    """
    logger.info(
        "Determining match for local service: {} and backend service: {}",
        local_service.Spec.Name,
        parent_service.name,
    )

    sync_report = SyncSurveyReport()

    # both are not none test
    if local_service is None or backend_service is None:
        logger.info("Not a match - one of the services is None")
        return sync_report
    logger.debug("both inputs are not None test passed")
    sync_report.both_inputs_are_not_none = True

    # image test
    if (
        local_service.Spec.TaskTemplate.ContainerSpec.Image
        != parent_service.requested_service_spec.image
    ):
        logger.info("Not a match - image mismatch")
        return sync_report
    logger.debug("image test passed")
    sync_report.image_in_sync = True

    # ports test
    requested_ports: List[Tuple[int, int]] = []
    for ports in parent_service.requested_service_spec.ports:
        try:
            target_port = int(ports.split(":")[0])
            published_port = int(ports.split(":")[1])
            requested_ports.append((target_port, published_port))
        except Exception as e:
            logger.error("Failed to parse ports. Value: {}, error: {}", ports, e)
            return sync_report
    for port in local_service.Endpoint.Spec.Ports:
        if (port.TargetPort, port.PublishedPort) not in requested_ports:
            logger.info("Not a match - ports mismatch")
            logger.debug("requested ports: {}", requested_ports)
            logger.debug(
                "local ports: {}",
                [
                    (port.TargetPort, port.PublishedPort)
                    for port in local_service.Endpoint.Spec.Ports
                ],
            )
            return sync_report
    local_ports = [
        (port.TargetPort, port.PublishedPort)
        for port in local_service.Endpoint.Spec.Ports
    ]
    for port in requested_ports:
        if port not in local_ports:
            logger.info("Not a match - ports mismatch")
            return sync_report
    logger.debug("ports test passed")
    sync_report.ports_in_sync = True

    # replicas test
    if (
        local_service.Spec.Mode.Replicated.Replicas
        != parent_service.requested_service_spec.replicas
    ):
        logger.info("Not a match - replicas mismatch")
        return sync_report
    logger.debug("replicas test passed")
    sync_report.replicas_in_sync = True

    # labels test
    requested_labels = (
        {
            label.split("=")[0]: label.split("=")[1]
            for label in parent_service.requested_service_spec.labels
        }
        if parent_service.requested_service_spec.labels
        else {}
    )
    if local_service.Spec.Labels != requested_labels:
        logger.info("Not a match - labels mismatch")
        return sync_report
    logger.debug("labels test passed")
    sync_report.labels_in_sync = True

    # env vars test
    local_env_vars = local_service.Spec.TaskTemplate.ContainerSpec.Env
    requested_env_vars = parent_service.requested_service_spec.env_vars
    # the bool checks are for comparing [] to None
    if local_env_vars != requested_env_vars and bool(local_env_vars) != bool(
        requested_env_vars
    ):
        logger.info("Not a match - env vars mismatch")
        logger.debug("local env vars: {}", local_env_vars)
        logger.debug("requested env vars: {}", requested_env_vars)
        return sync_report
    logger.debug("env vars test passed")
    sync_report.env_vars_in_sync = True

    # hostname test
    local_hostname = local_service.Spec.TaskTemplate.ContainerSpec.Hostname
    requested_hostname = parent_service.requested_service_spec.hostname
    if (
        local_hostname != requested_hostname
        and bool(local_hostname) != bool(requested_hostname)
        and requested_hostname is not None
    ):
        logger.info("Not a match - hostname mismatch")
        logger.debug("local hostname: {}", local_hostname)
        logger.debug("requested hostname: {}", requested_hostname)
        return sync_report
    logger.debug("hostname test passed")
    sync_report.hostname_in_sync = True

    # mounts test
    requested_mounts = parent_service.requested_service_spec.mounts
    local_mounts = [
        f"{mount.Source}:{mount.Target}:rw"
        for mount in local_service.Spec.TaskTemplate.ContainerSpec.Mounts
    ]
    if requested_mounts != local_mounts and bool(requested_mounts) != bool(
        local_mounts
    ):
        logger.info("Not a match - mounts mismatch")
        logger.debug("requested mounts: {}", requested_mounts)
        logger.debug("local mounts: {}", local_mounts)
        return sync_report
    logger.debug("mounts test passed")
    sync_report.mounts_in_sync = True

    # TODO: Network test, need to query every network from docker client
    # to find it's ID, and then use that to determine if the network is in sync

    logger.info("service {} is in sync", local_service.Spec.Name)
    return sync_report


async def handle_paired_mismatched(
    pairs: List[Pair], base_url: str, token: str, node_id: int
) -> None:
    """Handle the paired mismatched services.

    Args:
        pairs (List[Pair]): The paired mismatched services.
        base_url (str): The base URL of the API.
        token (str): The token to use.
        node_id (str): The ID of the node.
    """
    logger.info("Handling paired mismatched services ...")
    local_only = []
    backend_only = []
    for local_service, backend_service in pairs:
        if not local_service:
            backend_only.append((None, backend_service))
            continue
        if not backend_service:
            local_only.append((local_service, None))
            continue
        await handle_out_of_sync_services(
            [(local_service, backend_service)], base_url, token, node_id
        )

    await handle_local_only(local_only, node_id, base_url, token)
    await handle_backend_only(backend_only, base_url, token)


async def handle_out_of_sync_services(
    pairs: List[Pair], base_url: str, token: str, node_id: int
) -> None:
    """Handle the out of sync services.

    Args:
        pairs (List[Pair]): The out of sync services.
        base_url (str): The base URL
        token (str): The token to use
        node_id (int): The ID of the node
    """

    logger.info("Handling out of sync services ...")

    # update the docker attributes of the service node relationship
    # and change state to deployed and status to pending
    for local_service, backend_service in pairs:
        await update_service_node_relationship(
            base_url=base_url,
            token=token,
            service_node_relationship=ServiceNodeRelationship(
                service_id=backend_service.service_id,
                node_id=node_id,
                state=backend_service.state,
                status="pending",
                error="Service is out of sync",
                docker_service_attrs=local_service.model_dump(),
            ),
        )
        logger.info(
            "Updated service node relationship for service: {}",
            backend_service.service_id,
        )
    logger.info("Handled out of sync services.")


async def handle_paired_and_in_sync(
    pairs: List[Pair], base_url: str, token: str
) -> None:
    """Handle the in-sync services.

    Args:
        pairs (List[Pair]): The in-sync services.
        base_url (str): The base URL of the API.
        token (str): The token to use.
    """
    logger.info("Handling in-sync services ...")
    # update the docker attributes of the service node relationship
    # and change state to deployed and status to success
    for local_service, backend_service in pairs:
        await update_service_node_relationship(
            base_url=base_url,
            token=token,
            service_node_relationship=ServiceNodeRelationship(
                service_id=backend_service.service_id,
                node_id=backend_service.node_id,
                state="deployed",
                status="success",
                docker_service_attrs=local_service.model_dump(),
            ),
        )
        logger.info(
            "Updated service node relationship for service: {}",
            backend_service.service_id,
        )
    logger.info("Handled in-sync services.")


async def handle_local_only(
    pairs: List[Pair], node_id: int, base_url: str, token: str
) -> None:
    """Handle the local only services.

    Args:
        pairs (List[Pair]): The local only services.
        node_id (int): The ID of the node.
        base_url (str): The base URL of the API.
        token (str): The token to use.
    """
    logger.info("Handling local only services ...")
    logger.debug("Local only pairs: {}", pairs)
    # create out of sync services in the API
    for local_service, _ in pairs:
        logger.info("Creating out of sync service: {}", local_service.Spec.Name)
        local_mounts = local_service.Spec.TaskTemplate.ContainerSpec.Mounts
        new_service = await create_service(
            base_url=base_url,
            token=token,
            service_request=ServiceCreate(
                name=local_service.Spec.Name,
                resource_type="service",
                description=f"Service imported from Node {node_id}",
                desired_state="created",
                node_ids=[node_id],
                requested_service_spec=RequestedServiceSpec(
                    image=local_service.Spec.TaskTemplate.ContainerSpec.Image,
                    tags=["imported"],
                    ports=[
                        f"{port.TargetPort}:{port.PublishedPort}"
                        for port in local_service.Endpoint.Spec.Ports
                    ],
                    replicas=local_service.Spec.Mode.Replicated.Replicas,
                    labels=[
                        f"{key}={value}"
                        for key, value in local_service.Spec.Labels.items()
                    ],
                    env_vars=local_service.Spec.TaskTemplate.ContainerSpec.Env,
                    hostname=local_service.Spec.TaskTemplate.ContainerSpec.Hostname,
                    mounts=[
                        f"{mount.Type}:{mount.Source}:{mount.Target}"
                        for mount in local_mounts
                    ],
                ),
            ),
        )
        await update_service_node_relationship(
            base_url=base_url,
            token=token,
            service_node_relationship=ServiceNodeRelationship(
                service_id=new_service.id,
                node_id=node_id,
                state="created",
                status="success",
                docker_service_attrs=local_service.model_dump(),
            ),
        )


async def handle_backend_only(pairs: List[Pair], base_url: str, token: str) -> None:
    """Handle the backend only services.

    Args:
        pairs (List[Pair]): The backend only services.
        base_url (str): The base URL of the API.
        token (str): The token to use.
    """
    # mark the services as out of sync
    for _, backend_service in pairs:
        await update_service(
            base_url=base_url,
            token=token,
            service_id=backend_service.id,
            state="requested",
        )


async def update_service(base_url: str, token: str, service_id: int, **values) -> None:
    """Update the service with the given ID.

    Args:
        base_url (str): The base URL of the API.
        token (str): The token to use.
        service_id (int): The ID of the service to update.
        **values: The values to update.
    """
    logger.info("Updating service: {}", service_id)
    resp_text: Optional[str] = None
    url = f"{base_url}/services/{service_id}/"
    logger.debug("PATCH {}", url)
    headers = {"Authorization": f"Bearer {token}"}

    try:
        async with aiohttp.ClientSession() as session:
            response, resp_text = await make_api_request(
                session, url, headers, method="patch", data=values
            )
            logger.debug("Response: {}", resp_text)
            response.raise_for_status()
            logger.info("Updated service: {}", service_id)
            get_service_cached.cache_invalidate(service_id, base_url, token)
    except Exception as e:
        logger.error("Failed to update service: {}", resp_text if resp_text else e)


async def create_service(
    base_url: str, token: str, service_request: ServiceCreate
) -> Service:
    """Create a service with the given request.

    Args:
        base_url (str): The base URL of the API.
        token (str): The token to use.
        service_request (ServiceCreate): The service request.
    """
    logger.info("Creating service ...")
    resp_text: Optional[str] = None
    url = f"{base_url}/services/"
    logger.debug("POST {}", url)
    headers = {"Authorization": f"Bearer {token}"}

    async with aiohttp.ClientSession() as session:
        try:
            resp, resp_text = await make_api_request(
                session, url, headers, method="post", data=service_request.model_dump()
            )
            # resp_text = await resp.text()
            logger.debug("Response: {}", json.dumps(resp_text, indent=4))
            resp.raise_for_status()
            logger.info("Created service")
            return Service.model_validate_json(resp_text)
        except Exception as e:
            logger.error(
                "Failed to create service: {}",
                json.dumps(resp_text, indent=4) if resp_text else e,
            )


async def update_service_node_relationship(
    base_url: str, token: str, service_node_relationship: ServiceNodeRelationship
) -> None:
    """Update the service node relationship with the given request.

    Args:
        base_url (str): The base URL of the API.
        token (str): The token to use.
        service_node_relationship (ServiceNodeRelationship): The service
            node relationship.
    """
    logger.info("Updating service node relationship ...")
    resp_text: Optional[str] = None
    url = f"{base_url}/services/node_relationships"
    logger.debug("PUT {}", url)
    headers = {"Authorization": f"Bearer {token}"}

    async with aiohttp.ClientSession() as session:
        try:
            resp, resp_text = await make_api_request(
                session,
                url,
                headers,
                method="put",
                data=service_node_relationship.model_dump(),
            )
            logger.debug("Response: {}", json.dumps(resp_text, indent=4))
            resp.raise_for_status()
            logger.info("Updated service node relationship")
        except Exception as e:
            logger.error(
                "Failed to update service node relationship: {}",
                json.dumps(resp_text, indent=4) if resp_text else e,
            )
